function [nodes,uh]=fem_1d_solver(geom,problem_data,p,Ne,varargin)
% fem_1d_solver: solve -mu u''+sigma u=f in Omega
% with Dirichlet and/or Neumann boundary conditions
% by either P1-fem or P2-fem on a uniform grid.
%
% [nodes,uh]=fem_1d_solver(geom,problem_data,p,Ne)
%
% Input: geom: struct with fields:
%          geom.xa, geom.xb = end-points of Omega
%        problem_data: struct with coefficients and functions data
%          problem_data.mu =  mu (positive constant)
%          problem_data.sigma =  sigma (positive constant)
%          problem_data.f =f function handle @(x) 
%          problem_data.gd0= Dirichlet datum at left end-point
%          problem_data.gd1= Dirichlet datum at right end-point
%          problem_data.gn0= Neumann datum at left end-point
%          problem_data.gn1= Neumann datum at right end-point
%          problem_data.bc= 'xx' boundary conditions type:
%              problem_data.bc(1)='d' if Dirichlet at left
%              problem_data.bc(1)='n' if Neumann at left
%              problem_data.bc(2)='d' if Dirichlet at right
%              problem_data.bc(2)='n' if Neumann at right
%          p = local polynomial degree (1 or 2)
%          Ne =  number of elements of the partition
% optional input: degree = degree of exactness of quadrature formulas
%                  (default value =4)
%
% Output: nodes = column array with the nodes of the mesh
%         uh = column array of the numerical solution

%   Written by Paola Gervasio
%   $Date: 2020/04/26$

if nargin==4
  degree=4;
else
  degree=varargin{1};
end
% assemble the matrix and the rhs
[A,f,nodes,dof]=fem_1d_system(geom,problem_data,p,Ne,degree);

N=length(f); % dimension of the system
% to deal with the non-homegeneous Dirichlet conditions
% reduce the system to non-Dirichlet dofs
b=f;
if ~isempty(dof.dir)
b=b(dof.nondir)-A(dof.nondir,dof.dir)*b(dof.dir);
A=A(dof.nondir,dof.nondir);
end

% solve the linear system: cpmpute the dir-homogeneous solution
u0=A\b;

% assemble the solution: u=u0+Rg (homoegenous solution + lifting of the
% Dirichlet data)

uh=zeros(N,1);
uh(dof.nondir)=u0;
if ~isempty(dof.dir)
uh(dof.dir)=f(dof.dir);
end

end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% fem-P1 and fem-P2: assemble the matrix and the rhs
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [A,f,nodes,dof]=fem_1d_system(geom,problem_data,p,Ne,degree)
h=(geom.xb-geom.xa)/Ne; % mesh-size
N=p*Ne+1; % number of nodes
nodes=linspace(geom.xa,geom.xb,N)'; % nodes

nloc=p+1; % number of nodes inside each element
% connectivity matrix
conn=(1:nloc)'; nmax=nloc;
for ie=2:Ne
  conn=[conn,(nmax:nmax+p)'];
  nmax=nmax+p;
end

% mass matrix on the reference interval (0,1)
if p==1
Ml=[ 2,1; 1,2 ]/6;
elseif p==2
 Ml=[4, 2, -1;
  2, 16, 2;
  -1, 2,  4]/30;
end
 
% assemble the global mass matrix
M=zeros(N);
for ie=1:Ne
  M(conn(:,ie),conn(:,ie))=M(conn(:,ie),conn(:,ie))+Ml*h;
end


%M=diag(sum(M,2)); % lumped mass matrix


% stiffness matrix on the reference interval (0,1)
if p==1
  Kl=[1,-1;-1,1];
elseif p==2
Kl=[7, -8, 1;
  -8,16,-8;
  1,-8,7]/3;
end

% assemble the global stiffness matrix
 K=zeros(N);
for ie=1:Ne
  K(conn(:,ie),conn(:,ie))=K(conn(:,ie),conn(:,ie))+Kl/h;
end


% matrix A
A=problem_data.mu*K+problem_data.sigma*M;
A=sparse(A);

% right hand side

% compute quadrature nodes and weights and
% evaluate the basis functions at the quad nodes
vmesh.nodes=nodes;
vmesh.conn=conn;
vmesh.ne=Ne;
[basis,~,qf]=fem_1d_setting(vmesh,p,degree);

% assemble the r.h.s.
f=zeros(N,1);
for ie=1:Ne
  connl=conn(:,ie);
  ff=problem_data.f(qf.xq_ph(:,ie)); 
floc=basis.phi'*(ff.*qf.wq_ph(:,ie));
f(connl)=f(connl)+floc;
end

% f=M*problem_data.f(nodes); % assemble the rhs by using the mass matrix
% f=problem_data.f(nodes); % for dirac-delta data

% boundary condition at the left end-point
dof.dir=[];
dof.nondir=(1:N)';

if problem_data.bc(1)=='d'
  f(1)=problem_data.gd0; % dirichlet 
  dof.dir=[dof.dir;1];
  dof.nondir=setdiff(dof.nondir,1) ;
else
  f(1)=f(1)+problem_data.gn0; % Neumann 
end


% boundary condition at the right end-point
if problem_data.bc(2)=='d'
  f(N)=problem_data.gd1; % dirichlet 
  dof.dir=[dof.dir;N];
  dof.nondir=setdiff(dof.nondir,N) ;
else
  f(N)=f(N)+problem_data.gn1; % Neumann 
end


end


  